import copy
import numpy as np
import matplotlib.pyplot as plt
import time
from scipy.optimize import least_squares
from scipy.sparse import lil_matrix
from scipy.sparse import vstack

import rospy


# https://stackoverflow.com/a/24837438/1076564
def merge_dicts(dict1, dict2):
    """Recursively merges dict2 into dict1."""
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return dict2
    for k in dict2:
        if k in dict1:
            dict1[k] = merge_dicts(dict1[k], dict2[k])
        else:
            dict1[k] = dict2[k]
    return dict1


class FactorGraph(object):
    """Factorgraph implementation for 2D localization using relative motion measurements, absolute and relative
    markers."""

    def __init__(self, ma=(12.0, 40.0, 0.0), mr=(7.0, 9.0, 0.0), mr_gt=None, solver_options=None):
        # Trajectory estimate
        self.x = [[0.0, 0.0, 0.0]]
        """The best estimate of the robot trajectory in world coordinates. List Nx3."""
        self.x_odom = copy.copy(self.x)
        """Trajectory of the robot constructed just by integrating the input odom measurements. Should be very similar
        to odom frame pose. List Nx3."""
        self.x_tmp = np.zeros((3, 2))
        """Temporary variable used just to save a few memory allocations. Numpy array 3x2."""

        # Absolute marker
        self.ma = np.array([[ma[0]], [ma[1]], [ma[2]]])
        """Pose of the absolute marker (should be given as task input). Numpy array 3x1."""
        self.z_ma = [[np.nan, np.nan, np.nan]]
        """Measurements of the absolute marker pose (in current robot frame). This array should have the same number
        of elements as self.x. Each time instant where no measurement is available should contain NaNs. List Nx3."""
        self.c_ma = [0.0]
        """Costs for the absolute marker pose residuals. List N."""

        # Relative marker
        self.mr = np.array([[mr[0]], [mr[1]], [mr[2]]])  # initialization
        """Current best estimate of the relative marker pose. Numpy array 3x1."""
        self.mr_gt = None if mr_gt is None else np.array([[mr_gt[0]], [mr_gt[1]], [mr_gt[2]]])
        """Ground truth pose of the relative marker (can be given as task input). Numpy array 3x1."""
        self.z_mr = [[np.nan, np.nan, np.nan]]
        """Measurements of the relative marker pose (in current robot frame). This array should have the same number
        of elements as self.x. Each time instant where no measurement is available should contain NaNs. List Nx3."""
        self.c_mr = [0.0]
        """Costs for the relative marker pose residuals. List N."""
        self.seen_mr = False
        """Whether the relative marker has already been seen."""

        # Odom
        self.z_odom = []
        """Odometry measurements (relative motion in robot body frame). Each time instant contains a valid (non-NaN)
        measurement. List Nx3."""
        self.c_odom = []
        """Costs for odometry motion residuals. List N."""
        self.t_odom = []
        """Timestamps of odometry measurements (as :py:class:`rospy.Time`). List N."""

        # ICP
        self.z_icp = []
        """Relative motion estimated from ICP (expressed in robot body frame). This array should have the same number
        of elements as self.x. Each time instant where no measurement is available should contain NaNs. List Nx3."""
        self.c_icp = []
        """Costs for ICP motion residuals. List N."""

        # Ground truth odometry
        self.z_gt_odom = []
        """Ground truth odometry measurements (absolute poses in world frame). Each time instant where no measurement is
        available should contain NaNs. List Nx3."""

        # Solver
        # TODO: ARO homework 3: test with various combinations of loss and f_scale, select the best
        # (loss, f_scale) tested configs: ('linear', 1.0), ('soft_l1', 0.05), ('huber', 1.0), ('cauchy', 1.9)
        self.solver_options = {
            "method": 'trf',
            "x_scale": 1,
            "tr_solver": 'lsmr',
            "tr_options": {
                'maxiter': 10,
                'atol': 1e-4,
                'btol': 1e-4
            },
            "loss": "soft_l1",
            "f_scale": 0.05,
            "max_nfev": 40,
            "verbose": 1,
            "ftol": 1e-3,  # /x_ini.shape[0],
            "gtol": 1e-8,
            "xtol": 1e-4,  # /x_ini.shape[0],
        }
        if solver_options is not None:
            merge_dicts(self.solver_options, solver_options)
        rospy.loginfo("Solver options are: " + str(self.solver_options))

        self._figure_shown = False

    def find_nearest_index_for_new_message(self, stamp=None):
        """Find the best index into self.x (or self.z_*) for a measurement with the given timestamp.

        :param stamp: The timestamp. If `None`, the measurement is inserted to the latest time.
        :type stamp: rospy.Time or None
        :return: The index (using Python array indexing, e.g. it can be -1 to signify last element). `None` is returned
                 if there are no odom measurements yet and thus there is no place to put the measurement.
        :rtype: int or None
        """
        # no odom measurements yet -> we don't have anywhere to put the message
        if len(self.t_odom) == 0:
            return None

        # no timestamp -> always add to the last position
        # only one odom measurement -> add to the only position
        if stamp is None or len(self.t_odom) <= 1:
            return -1

        # timestamp is newer than the newest odom measurement -> add to the last position
        if stamp > self.t_odom[-1]:
            return -1

        # from now on, we have at least two measurements, and timestamp is not newer than the newest odom measurement

        # we perform a linear search from the end of the list to find the best slot
        # this is generally not very good for performance, but we assume we almost always find a slot near the end,
        # so it would not be worth it to implement binary search here
        i = len(self.t_odom) - 1
        while stamp < self.t_odom[i] and i > 0:
            i -= 1

        # i now points to the first slot where timestamp is newer than the measurement
        # so i + 1 is the first slot where timestamp is older than the measurement
        diff_older = abs(stamp - self.t_odom[i])
        diff_newer = abs(stamp - self.t_odom[i + 1])

        # choose the closer slot
        if diff_older < diff_newer:
            return i
        return i + 1

    def add_z(self, z_odom, c_odom, z_mr, c_mr, z_ma, c_ma, z_icp, c_icp, z_gt_odom,
              t_odom=None, t_markers=None, t_icp=None):
        """Add measurements to the factorgraph.

        :param tuple z_odom: Odometry relative motion measurement. Should not contain NaNs. List 3.
        :param float c_odom: Cost of the odometry relative motion residual.
        :param tuple z_mr: Measurement of the relative marker pose in robot body frame (NaNs if not observed). List 3.
        :param float c_mr: Cost of the relative marker pose measurement residual.
        :param tuple z_ma: Measurement of the absolute marker pose in robot body frame (NaNs if not observed). List 3.
        :param float c_ma: Cost of the absolute marker pose measurement residual.
        :param tuple z_icp: Measurement of velocity from ICP in robot body frame (NaNs if not observed). List 3.
        :param float c_icp: Cost of the ICP velocity measurement residual.
        :param tuple z_gt_odom: Measurement of ground truth absolute pose in world frame (NaNs if not observed). List 3.
        :param t_odom: Timestamp of the odometry measurement.
        :type t_odom: rospy.Time or None
        :param t_markers: Timestamp of the marker observations.
        :type t_markers: rospy.Time or None
        :param t_icp: Timestamp of the ICP measurement.
        :type t_icp: rospy.Time or None
        """
        # Add odometry relative motion measurement
        self.z_odom.append(z_odom)
        self.c_odom.append(c_odom)
        self.t_odom.append(t_odom)

        # Add relative and absolute marker measurements
        self.z_ma.append(np.nan * np.zeros_like(z_ma))
        self.c_ma.append(c_ma)
        self.z_mr.append(np.nan * np.zeros_like(z_mr))
        self.c_mr.append(c_mr)
        init_mr = False

        markers_i = self.find_nearest_index_for_new_message(t_markers)
        if markers_i is not None:
            if not np.isnan(z_ma[0]):
                self.z_ma[markers_i] = z_ma

            if not np.isnan(z_mr[0]):
                self.z_mr[markers_i] = z_mr
                if not self.seen_mr:
                    init_mr = True
                self.seen_mr = True

        # Add ICP relative motion measurement
        self.z_icp.append(np.nan * np.zeros_like(z_icp))
        self.c_icp.append(c_icp)

        icp_i = self.find_nearest_index_for_new_message(t_icp)
        if icp_i is not None and not np.isnan(z_icp[0]):
            self.z_icp[icp_i] = z_icp

        self.z_gt_odom.append(z_gt_odom)

        # Update self.x_odom by integrating odometry velocity
        z_odom_np = np.atleast_2d(z_odom).transpose()
        self.x_tmp[:, 0] = self.x_odom[-1]
        x, y, w = self.forward_kinematics(self.x_tmp, z_odom_np)
        self.x_odom.append(np.array([x, y, w]).ravel().tolist())

        # Initialize new x by integrating odom
        self.x_tmp[:, 0] = self.x[-1]
        x, y, w = self.get_world_pose(self.x_tmp, z_odom_np)
        self.x.append(np.array([x, y, w]).ravel().tolist())

        # If this is the first time the relative marker has been seen, set its pose estimate to the observed one
        if init_mr:
            self.mr = self.get_world_pose(np.array(z_mr)[:, np.newaxis], np.array(self.x[-1])[:, np.newaxis])

    def visu(self, x, mr, ma, z_odom, z_mr, z_ma, z_gt_odom):
        x = x.transpose()
        x_odom = np.array(self.x_odom).transpose()

        idx_mr = np.where((~np.isnan(z_mr)).all(axis=0))[0]
        idx_ma = np.where((~np.isnan(z_ma)).all(axis=0))[0]
        idx_gt = np.where((~np.isnan(z_gt_odom)).all(axis=0))[0]

        z_mr_wcf = self.get_world_pose(z_mr[:, idx_mr], x[:, idx_mr])
        z_ma_wcf = self.get_world_pose(z_ma[:, idx_ma], x[:, idx_ma])
        z_icp = np.array(self.z_icp).transpose()
        gt_odom = np.array(z_gt_odom)

        nplots = 2  # update this when more subplots are added
        if idx_gt.shape[0] > 0:
            nplots += 2  # update this when more GT subplots are added
        plot_index = 0

        fig = plt.figure(1, figsize=(10, 13))
        plt.clf()

        plot_index += 1
        main_ax = plt.subplot(nplots, 1, (plot_index, plot_index + 1))
        plot_index += 1
        for k, j in enumerate(idx_ma):
            main_ax.plot([x[0, j], z_ma_wcf[0, k]], [x[1, j], z_ma_wcf[1, k]], 'x-', color='k', linewidth=1, mew=2)
            main_ax.plot([ma[0, 0], z_ma_wcf[0, k]], [ma[1, 0], z_ma_wcf[1, k]], ':', color='k', linewidth=1, mew=2)

        for k, j in enumerate(idx_mr):
            main_ax.plot([x[0, j], z_mr_wcf[0, k]], [x[1, j], z_mr_wcf[1, k]], 'x-', color='c', linewidth=1, mew=2)
            main_ax.plot([mr[0, 0], z_mr_wcf[0, k]], [mr[1, 0], z_mr_wcf[1, k]], ':', color='c', linewidth=1, mew=2)

        if idx_gt.shape[0] > 0:
            main_ax.plot(gt_odom[0, :], gt_odom[1, :], '-', color='r', linewidth=1, mew=2, label="GT odom")
        main_ax.plot(x_odom[0, :], x_odom[1, :], '-', color='y', linewidth=1, mew=2, label="Integrated odom")
        main_ax.plot(z_icp[0, :], z_icp[1, :], '-', color='m', linewidth=1, mew=2, label="ICP odom")
        main_ax.plot(x[0, :], x[1, :], '-', color='b', linewidth=2, mew=2, label="Fused odom")
        if self.mr_gt is not None:
            main_ax.plot(self.mr_gt[0, 0], self.mr_gt[1, 0], 's', color='y', linewidth=1, mew=6, label="Rel.m. GT")
        main_ax.plot(mr[0, :], mr[1, :], 's', color='c', linewidth=1, mew=6, label="Rel. marker")
        main_ax.plot(ma[0, :], ma[1, :], 's', color='k', linewidth=1, mew=6, label="Abs. marker")

        main_ax.axis('equal')
        main_ax.set_xlabel('x')
        main_ax.set_ylabel('y')
        main_ax.grid()
        main_ax.legend()

        if idx_gt.shape[0] > 0:
            plot_index += 1
            gt_ax = plt.subplot(nplots, 1, plot_index)
            errors_x = x[0, 1:] - gt_odom[0, :]
            errors_y = x[1, 1:] - gt_odom[1, :]
            errors_yaw = (x[2, 1:] - gt_odom[2, :] + np.pi) % (2 * np.pi) - np.pi
            errors = np.linalg.norm(np.vstack((errors_x, errors_y, errors_yaw)), axis=0)
            stamps = [(t - self.t_odom[0]).to_sec() for t in self.t_odom[1:x.shape[1]]]
            gt_ax.plot(stamps, errors, label="FG localization error", color='k')
            gt_ax.plot(stamps, errors_x, label="X error", color='r')
            gt_ax.plot(stamps, errors_y, label="Y error", color='g')
            gt_ax.plot(stamps, errors_yaw, label="Yaw error", color='b')
            gt_ax.set_xlabel('t')
            gt_ax.set_ylabel('fused error')
            gt_ax.set_ylim(min(np.min(errors_x), np.min(errors_y), np.min(errors_yaw), 0) * 1.1,
                           max(np.max(errors), np.max(errors_x), np.max(errors_y), np.max(errors_yaw)) * 1.1)
            gt_ax.grid()
            gt_ax.legend()

            plot_index += 1
            gt_odom_ax = plt.subplot(nplots, 1, plot_index)
            odom_and_gt_idxs = np.where((~np.isnan(x_odom[:, 1:]) & ~np.isnan(gt_odom)).all(axis=0))[0]
            if odom_and_gt_idxs.shape[0] > 0:
                # TODO this needs more debugging
                first_odom_idx = odom_and_gt_idxs[0]
                first_odom = x_odom[:, first_odom_idx + 1]
                first_gt = gt_odom[:, first_odom_idx]
                yaw_diff = first_gt[2] - first_odom[2]
                R = np.array([[np.cos(yaw_diff), -np.sin(yaw_diff)], [np.sin(yaw_diff), np.cos(yaw_diff)]])
                x_odom_world = np.zeros_like(x_odom)
                x_odom_world[0:2, :] = np.matmul(R, x_odom[0:2, :]) + first_gt[0:2, np.newaxis] - first_odom[0:2, np.newaxis]
                x_odom_world[2, :] = x_odom_world[2, :] + yaw_diff
                print(first_gt)
                print(first_odom)
                print(x_odom_world[:, 0])
                print(x_odom[:, 0])
                odom_errors_x = x_odom_world[0, 1:] - gt_odom[0, :]
                odom_errors_y = x_odom_world[1, 1:] - gt_odom[1, :]
                odom_errors_yaw = (x_odom_world[2, 1:] - gt_odom[2, :] + np.pi) % (2 * np.pi) - np.pi
                odom_errors = np.linalg.norm(np.vstack((odom_errors_x, odom_errors_y, odom_errors_yaw)), axis=0)
                gt_odom_ax.plot(stamps, odom_errors, label="Odom localization error", color='m')
                gt_odom_ax.plot(stamps, odom_errors_x, label="X error", color='r')
                gt_odom_ax.plot(stamps, odom_errors_y, label="Y error", color='g')
                gt_odom_ax.plot(stamps, odom_errors_yaw, label="Yaw error", color='b')
                gt_odom_ax.set_ylim(min(np.min(odom_errors_x), np.min(odom_errors_y), np.min(odom_errors_yaw), 0) * 1.1,
                                    max(np.max(odom_errors), np.max(odom_errors_x), np.max(odom_errors_y),
                                        np.max(odom_errors_yaw)) * 1.1)
            gt_odom_ax.set_xlabel('t')
            gt_odom_ax.set_ylabel('odom error')
            gt_odom_ax.grid()
            gt_odom_ax.legend()

        plt.savefig("/tmp/fig%04i.png" % (x.shape[1],))

        # plt.pause() brings the plot window into foreground on each redraw, which is not convenient, so we only call it
        # the first time and then we use a workaround that hopefully works on all plotting backends (but not sure!)
        if not self._figure_shown:
            plt.pause(0.001)
            self._figure_shown = True
        else:
            fig.canvas.draw_idle()
            fig.canvas.start_event_loop(0.001)

    def optimize(self, x, mr, ma, z_odom, z_mr, z_ma, z_icp, c_odom, c_mr, c_ma, c_icp):
        """Optimize the factor graph given by the parameters.

        :param np.ndarray x: Robot trajectory estimate. Numpy array Nx3.
        :param np.ndarray mr: Relative marker pose estimate. Numpy array 3x1.
        :param np.ndarray ma: Absolute marker pose. Numpy array 3x1.
        :param np.ndarray z_odom: Odometry measurements (relative motion in robot body frame). Numpy array 3xN.
        :param np.ndarray z_mr: Relative marker pose measurements (in robot body frame). Numpy array 3xN.
        :param np.ndarray z_ma: Absolute marker pose measurements (in robot body frame). Numpy array 3xN.
        :param np.ndarray z_icp: ICP odometry measurements (relative motion in robot body frame). Numpy array 3xN.
        :param np.ndarray c_odom: Costs of odometry residuals. Numpy array N.
        :param np.ndarray c_mr: Costs of relative marker pose estimate residuals. Numpy array N.
        :param np.ndarray c_ma: Costs of absolute marker pose estimate residuals. Numpy array N.
        :param np.ndarray c_icp: Costs of ICP odometry residuals. Numpy array N.
        :return: Optimized `x` and `mr` (robot trajectory estimate and relative marker pose estimate).
        """
        x_ini = np.hstack((x.reshape(-1), mr.reshape(-1)))   # concatenation [x, mr]
        idx_ma = np.where((~np.isnan(z_ma)).all(axis=0))[0]
        idx_mr = np.where((~np.isnan(z_mr)).all(axis=0))[0]
        idx_icp = np.where((~np.isnan(z_icp)).all(axis=0))[0]

        sol = least_squares(self.compute_residuals, x_ini, self.compute_residuals_jacobian,
                            args=(ma, z_odom, z_mr, z_ma, z_icp, idx_mr, idx_ma, idx_icp, c_odom, c_mr, c_ma, c_icp),
                            **self.solver_options)

        # "un-concatenate" x, mr
        x, mr = sol.x[0:-3].reshape(x.shape[0], x.shape[1]).tolist(), sol.x[-3:].reshape(3, 1)
        return x, mr

    def get_world_pose(self, z_m, x):
        """Get world pose of the given measurements in robot body frame.

        :param np.ndarray z_m: The measurements in robot body frame. Numpy array 3xN.
        :param np.ndarray x: Robot world pose estimate in times corresponding to each measurement. Numpy array 3XN.
        :return: World poses of the measurements computed using the robot pose estimates. Numpy array 3XN.
        :rtype: np.ndarray
        """
        x_t = np.zeros(x.shape)
        for t in range(x.shape[1]):
            R = np.array([[np.cos(x[2, t]), -np.sin(x[2, t])], [np.sin(x[2, t]), np.cos(x[2, t])]])  # rotation matrix
            x_t[0:2, t] = np.matmul(R, z_m[0:2, t]) + x[0:2, t]
            x_t[2, t] = z_m[2, t] + x[2, t]
        return x_t

    def forward_kinematics(self, x, z_odom):
        """Forward kinematics model of the robot motion.

        :param numpy.ndarray x: Robot pose estimates in world frame. The last column is ignored. Numpy array 3x(N+1).
        :param numpy.ndarray z_odom: Odometry measurements (relative motion in robot body frame). Numpy array 3xN.
        :return: Robot poses with one step of forward kinematics applied to them. Numpy array 3xN.
        :rtype: np.ndarray
        """
        cos_yaw = np.cos(x[2, 0:-1])
        sin_yaw = np.sin(x[2, 0:-1])
        pos_x = x[0, 0:-1] + z_odom[0, :] * cos_yaw + z_odom[1, :] * sin_yaw
        pos_y = x[1, 0:-1] + z_odom[0, :] * sin_yaw + z_odom[1, :] * cos_yaw
        pos_w = x[2, 0:-1] + z_odom[2, :]
        return pos_x, pos_y, pos_w

    def compute_residuals(self, xmr, ma, z_odom, z_mr, z_ma, z_icp, idx_mr, idx_ma, idx_icp, c_odom, c_mr, c_ma, c_icp):
        """Compute residuals (errors) of all measurements.

        :param np.ndarray xmr: Concatenated and linearized x and mr. Numpy array (3N+3)x1.
        :param np.ndarray ma: Absolute marker pose. Numpy array 3x1.
        :param np.ndarray z_odom: Odometry measurements (relative motion in robot body frame). Numpy array 3xN.
        :param np.ndarray z_mr: Relative marker pose measurements (in robot body frame). Numpy array 3xN.
        :param np.ndarray z_ma: Absolute marker pose measurements (in robot body frame). Numpy array 3xN.
        :param np.ndarray z_icp: ICP odometry measurements (relative motion in robot body frame). Numpy array 3xN.
        :param np.ndarray idx_mr: Indices of z_mr with valid (non-NaN) measurements. Numpy array N.
        :param np.ndarray idx_ma: Indices of z_ma with valid (non-NaN) measurements. Numpy array N.
        :param np.ndarray idx_icp: Indices of z_icp with valid (non-NaN) measurements. Numpy array N.
        :param np.ndarray c_odom: Costs of odometry residuals. Numpy array N.
        :param np.ndarray c_mr: Costs of relative marker pose estimate residuals. Numpy array N.
        :param np.ndarray c_ma: Costs of absolute marker pose estimate residuals. Numpy array N.
        :param np.ndarray c_icp: Costs of ICP odometry residuals. Numpy array N.
        :return: The concatenated and linearized residuals. Numpy array Qx1, Q <= 9N.
        :rtype: np.ndarray
        """
        # unpack x, mr from the concatenated state
        x, mr = xmr[0:-3].reshape(3, int((xmr.shape[0] - 3) / 3), order='F'), xmr[-3:].reshape(3, 1)
        odom_x, odom_y, odom_w = self.forward_kinematics(x, z_odom)

        # Robot trajectory computed by taking previous pose and z_odom measurement and propagating them via the
        # forward kinematics model.
        observed_x = np.vstack((odom_x[None, :], odom_y[None, :], odom_w[None, :]))

        # Odom measurements residuals
        # TODO: ARO homework 3: compute odometry residuals in res_odom
        res_odom = np.zeros_like(x[:, 1:])
        res_odom = np.repeat(np.atleast_2d(c_odom), res_odom.shape[0], axis=0) * res_odom  # apply cost

        # Relative marker measurements residuals
        observed_mr = self.get_world_pose(z_mr[:, idx_mr], x[:, idx_mr])
        # TODO: ARO homework 3: compute relative marker residuals in res_mr
        res_mr = np.zeros_like(observed_mr)
        res_mr[2, :] = np.mod(res_mr[2, :] + np.pi, np.pi * 2) - np.pi  # wrap around pi
        res_mr = np.repeat(np.atleast_2d(c_mr)[:, idx_mr], res_mr.shape[0], axis=0) * res_mr  # apply cost

        # Absolute marker measurements residuals
        observed_ma = self.get_world_pose(z_ma[:, idx_ma], x[:, idx_ma])
        # TODO: ARO homework 3: compute absolute marker residuals in res_ma
        res_ma = np.zeros_like(observed_ma)
        res_ma[2, :] = np.mod(res_ma[2, :] + np.pi, np.pi * 2) - np.pi  # wrap around pi
        res_ma = np.repeat(np.atleast_2d(c_ma)[:, idx_ma], res_ma.shape[0], axis=0) * res_ma  # apply cost

        # TODO: after homework 4: integrate ICP odometry

        # Concatenate and linearize the residuals
        return np.hstack([
            res_odom.reshape(-1, order='F'),
            res_mr.reshape(-1, order='F'),
            res_ma.reshape(-1, order='F')])

    def compute_residuals_jacobian(
            self, xmr, ma, z_odom, z_mr, z_ma, z_icp, idx_mr, idx_ma, idx_icp, c_odom, c_mr, c_ma, c_icp):
        """Compute Jacobian of the residuals. This is used to steer the optimization in a good direction.

        :param np.ndarray xmr: Concatenated and linearized x and mr. Numpy array (3N+3)x1.
        :param np.ndarray ma: Absolute marker pose. Numpy array 3x1.
        :param np.ndarray z_odom: Odometry measurements (relative motion in robot body frame). Numpy array 3xN.
        :param np.ndarray z_mr: Relative marker pose measurements (in robot body frame). Numpy array 3xN.
        :param np.ndarray z_ma: Absolute marker pose measurements (in robot body frame). Numpy array 3xN.
        :param np.ndarray z_icp: ICP odometry measurements (relative motion in robot body frame). Numpy array 3xN.
        :param np.ndarray idx_mr: Indices of z_mr with valid (non-NaN) measurements. Numpy array N.
        :param np.ndarray idx_ma: Indices of z_ma with valid (non-NaN) measurements. Numpy array N.
        :param np.ndarray idx_icp: Indices of z_icp with valid (non-NaN) measurements. Numpy array N.
        :param np.ndarray c_odom: Costs of odometry residuals. Numpy array N.
        :param np.ndarray c_mr: Costs of relative marker pose estimate residuals. Numpy array N.
        :param np.ndarray c_ma: Costs of absolute marker pose estimate residuals. Numpy array N.
        :param np.ndarray c_icp: Costs of ICP odometry residuals. Numpy array N.
        :return: Jacobian of the residuals. Numpy array Qx(3N+3), Q <= 9N.
        :rtype: np.ndarray
        """
        # unpack x, mr from the concatenated state
        x, mr = xmr[0:-3].reshape(3, int((xmr.shape[0] - 3) / 3), order='F'), xmr[-3:].reshape(3, 1)
        num_positions, dim, num_rel_markers, num_abs_markers = x.shape[1], x.shape[0], mr.shape[1], ma.shape[1]

        # lil_matrix is a sparse matrix implementation. We use it here to speed up the computations.

        # Odom measurements
        J_odom = lil_matrix((dim * (num_positions - 1), dim * (num_positions + num_rel_markers)), dtype=np.float32)
        for t in range(num_positions - 1):
            # TODO: ARO homework 3: compute the odom part of the Jacobian in J, i.e.
            #       differentiate res_odom[:, t] w.r.t x
            J = np.array((
                (1, 0, 0),
                (0, 1, 0),
                (0, 0, 1)))
            J = c_odom[t] * J  # apply cost
            j = t * dim  # index of t-th pose in xmr
            J_odom[j:(j + 3), j:(j + 6)] = np.hstack((J, -np.eye(3)))

        # Relative marker measurements
        J_mr = lil_matrix((dim * idx_mr.shape[0], dim * (num_positions + num_rel_markers)), dtype=np.float32)
        k = 0
        for t in idx_mr:
            # TODO: ARO homework 3: compute the relative marker part of the Jacobian in J, i.e.
            #       differentiate res_mr[:, t] w.r.t x
            J = np.array((
                (1, 0, 0),
                (0, 1, 0),
                (0, 0, 1)))
            J = c_mr[t] * J  # apply cost
            j = t * dim  # index of t-th pose in xmr
            J_mr[k:(k + 3), j:(j + 3)] = J
            J_mr[k:(k + 3), dim * num_positions:] = -np.eye(3)  # derivative of -mr w.r.t mr
            k = k + 3

        # Absolute marker measurements
        J_ma = lil_matrix((dim * idx_ma.shape[0], dim * (num_positions + num_abs_markers)), dtype=np.float32)
        k = 0
        for t in idx_ma:
            # TODO: ARO homework 3: compute the absolute marker part of the Jacobian in J, i.e.
            #       differentiate res_ma[:, t] w.r.t x
            J = np.array((
                (1, 0, 0),
                (0, 1, 0),
                (0, 0, 1)))
            J = c_ma[t] * J  # apply cost
            j = t * dim  # index of t-th pose in xmr
            J_ma[k:(k + 3), j:(j + 3)] = J
            k = k + 3

        # TODO: after homework 4: integrate ICP odometry

        return vstack((J_odom, J_mr, J_ma))
